////////////////////////////////////////////////////////////////////////////////
// PushBlock a java game. Try to move the blue spheres to a read shaded area.
//
// Copyright (C) 1999 Anders Reggestad andersr@pvv.ntnu.no
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
////////////////////////////////////////////////////////////////////////////////

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.util.*;

public class PushBlock extends Applet 
implements Runnable, KeyListener, MouseListener
{
  // Thread used to load images
  Thread	 m_PushBlock = null;

  static final int LOADING = 0;
  static final int INTRO   = 1;
  static final int PLAYING = 2;
  static final int SCORE   = 3;

  // Array that hold all images used by this Applet!
  Image     m_images[];
  int       m_imgWidth  = 0;
  int       m_imgHeight = 0;
  boolean   m_fAllLoaded = false;
  boolean   m_fIntroLoaded = false;
  int       m_state = LOADING;
  int       m_level = 0;
  Image     m_imgIntro;
  Image     m_doubleBuffer = null;

  static final String  m_imageNames[] =
    {"outside.gif","floor.gif","wall.gif","goal.gif",
     "blue_sphere.gif","player.gif","water.gif","dirt.gif","floor2.gif"};
  final int NUM_IMAGES = 9;

  public static final String MSG_LOADING  = "Loading images...";
  public static final String MSG_SCORE    = "Your score is: ";
  public static final String MSG_MOVES    = "Moves : ";
  public static final String MSG_BOARD    = "Level info : ";
  public static final String MSG_NEWGAME  = "Press space for new game!";
  public static final String MSG_NEXTGAME = "Press space for next game!";

  PushBlockBoard     m_board;

  public void init() {
    m_state = LOADING;
    resize(288, 288);
    m_board = newBoard(m_level);
    m_doubleBuffer = createImage(288,288);
    addKeyListener(this);
    addMouseListener(this);
  }

  PushBlockBoard loadLevelFromURL(int level){
    PushBlockBoard board = null;
    try {
      URL levelURL = new URL(getDocumentBase(),"levels/pushblock_"+level+".lvl");
      BufferedReader in = new BufferedReader (
                              new InputStreamReader ( levelURL.openStream()));
      String inputLine;

      int state = 0,width=0,height,pieceIndex=0,line=0;
      while ((inputLine = in.readLine()) != null){
        System.out.println(inputLine);
        StringTokenizer st = new StringTokenizer(inputLine);
        switch (state){
        case 0: // Read config
          try {
            int version = Integer.parseInt(st.nextToken());
            width = Integer.parseInt(st.nextToken());
            height = Integer.parseInt(st.nextToken());
            board  = new PushBlockBoard(width,height,level,level+1);
            board.setImageSize(m_imgWidth,m_imgHeight);
            state = 1;
          } catch (NoSuchElementException e){
            board = newBoard(-1);
            return board;
          }
          break;
        case 1: // Read board
          board.setInfo(inputLine);
          state = 2;
          break;
        case 2: // Read board
          try {
            for (int ii = 0; ii < width ; ii++){
              String tile = st.nextToken();
              for (int jj = 0; jj < tile.length(); jj++){
                switch (tile.charAt(jj)){
                case '.': board.setTile(ii,line,PushBlockBoard.FLOOR); break;
                case '#': board.setTile(ii,line,PushBlockBoard.WALL); break;
                case 'x': board.setTile(ii,line,PushBlockBoard.GOAL); break;
                case '~': board.setTile(ii,line,PushBlockBoard.WATER); break;
                case ',': board.setTile(ii,line,PushBlockBoard.FRICTION); break;
                case 'o': 
                  board.addPiece(ii,line,PushBlockPiece.SPHERE);
                  break;
                case '*': 
                  board.addPiece(ii,line,PushBlockPiece.DIRT);
                  break;
                case '+':
                  board.setPosition(ii,line);
                  break;
                default: board.setTile(ii,line,PushBlockBoard.FLOOR); break;
                }
              }
            }
            line++;
          } catch (NoSuchElementException e){
            return board;
          }
        }
      }
      in.close();
  
    } catch (MalformedURLException e) {  // new URL() failed
      board = newBoard(-1);
    } catch (IOException e) { // Failed to connect to or read URL
      if (level != 0){
        board = loadLevelFromURL(0);
      } else board = newBoard(-1);
    }
    return board;
  }

  PushBlockBoard newBoard(int level){
    PushBlockBoard board;
    switch (level){
    case -1:  // Create default board
      board  = new PushBlockBoard(4,4,-1,-1);
      board.setImageSize(m_imgWidth,m_imgHeight);
      board.setInfo("Author: Anders Reggestad (andersr@pvv.ntnu.no)");
      board.setPosition(1,1);
      board.setTile(2,2,PushBlockBoard.GOAL);
      board.setTile(0,2,PushBlockBoard.WATER);
      board.setTile(1,1,PushBlockBoard.FRICTION);
      board.addPiece(1,2,PushBlockPiece.SPHERE);
      board.addPiece(2,1,PushBlockPiece.DIRT);
      break;
    default:
      board = loadLevelFromURL(level);
      break;
    }
    return board;
  }

  public void start(){
    if (m_PushBlock == null){
      m_PushBlock = new Thread(this);
      m_PushBlock.start();
    }
  }
	
  public void stop(){
    if (m_PushBlock != null){
      m_PushBlock.stop();
      m_PushBlock = null;
    }
  }

  public void destroy(){
  }

  public String getAppletInfo(){
    return "Name: PushBlock\r\n" +
           "Author: Anders Reggestad\r\n"+
           "Copyright: 1999 andersr@pvv.ntnu.no\r\n";
  }

  public void keyTyped(KeyEvent e){
  }

  public void keyPressed(KeyEvent e){
  }

  void clicked(){
    if (m_fAllLoaded){
      switch(m_state){
      case INTRO:
        m_state = PLAYING;
        break;
      case SCORE:
        m_level = m_board.getNextLevel();
        m_board = newBoard(m_level);
        m_state = PLAYING;
        break;
      }
      repaint();
    }
  }

  public void keyReleased(KeyEvent e){
    switch (e.getKeyCode()){
    case KeyEvent.VK_SPACE:
    case KeyEvent.VK_ENTER:
      clicked();
      break;
    case KeyEvent.VK_R: // Reload level!!
      if (m_state == PLAYING){
        m_board = newBoard(m_level);
        m_state = PLAYING;
      }
      break;
    case KeyEvent.VK_UP:
      if (m_state == PLAYING)
        m_board.moveUp();
      break;
    case KeyEvent.VK_DOWN:
      if (m_state == PLAYING)
        m_board.moveDown();
      break;
    case KeyEvent.VK_LEFT:
      if (m_state == PLAYING)
        m_board.moveLeft();
      break;
    case KeyEvent.VK_RIGHT:
      if (m_state == PLAYING)
        m_board.moveRight();
      break;
    }
    if ((m_state == PLAYING) && m_board.isFinished()){
      m_state = SCORE;
      m_level++;
    }
    repaint();
  }

  public void mouseClicked(MouseEvent e){
      clicked();
  }

  public void mousePressed(MouseEvent e){
  }

  public void mouseReleased(MouseEvent e){
  }

  public void mouseEntered(MouseEvent e){
  }

  public void mouseExited(MouseEvent e){
  }

  public void paint(Graphics screenGraphics){
    Graphics g = m_doubleBuffer.getGraphics();
    g.setColor(Color.white);
    g.fillRect(0,0,288,288);
    switch (m_state){
    case LOADING:
      g.setColor(Color.red);
      g.drawString(MSG_LOADING, 16, 20);
      break;
    case INTRO:
      g.setColor(Color.red);
      if (!m_fAllLoaded){
        g.drawString(MSG_LOADING, 16, 20);
      } else {
        g.drawString(MSG_NEWGAME, 16, 20);
      }
      if (m_fIntroLoaded){
        g.drawImage(m_imgIntro,0,0,null);
      }
      break;
    case PLAYING:
      if (m_fAllLoaded){
        m_board.paint(g,m_images);
        g.setColor(Color.yellow);
	if (m_board.getScore()!=0){
          g.drawString(MSG_MOVES + m_board.getScore(), 16, 276);
        } else {
          g.drawString(MSG_BOARD, 16, 244);
          g.drawString(m_board.getInfo(), 16, 276);
        }
      } 
      break;
    case SCORE:
      if (m_fAllLoaded){
        m_board.paint(g,m_images);
        g.setColor(Color.yellow);
        g.drawString(MSG_SCORE + m_board.getScore(), 16, 20);
        g.drawString(MSG_NEXTGAME, 16, 52);
      } 
      break;
    }
    screenGraphics.drawImage(m_doubleBuffer,0,0,null);
  }

  public void update(Graphics g){
    paint(g);
  }

  public void run(){
		
    // If re-entering the page, then the images have already been loaded.
    // m_fAllLoaded == TRUE.
    if (!m_fAllLoaded){
      repaint();
      m_images   = new Image[NUM_IMAGES];

      // Load intro image
      MediaTracker introTracker = new MediaTracker(this);
      m_imgIntro = getImage(getDocumentBase(), "images/intro.gif" );
      introTracker.addImage(m_imgIntro, 0);
      try {
        introTracker.waitForAll();
        m_fIntroLoaded = !introTracker.isErrorAny();
        m_state = INTRO;
        repaint();
      }	catch (InterruptedException e) {
      }

      // Load in all the images
      MediaTracker tracker = new MediaTracker(this);
      String strImage;

      // For each image in the animation, this method first constructs a
      // string containing the path to the image file; then it begins
      // loading the image into the m_Images array.  Note that the call to
      // getImage will return before the image is completely loaded.
      for (int i = 0; i < NUM_IMAGES; i++){

        // Build path to next image
        strImage = "images/" + m_imageNames[i];
        m_images[i] = getImage(getDocumentBase(), strImage);

        tracker.addImage(m_images[i], 0);
      }

      // Wait until all images are fully loaded
      try {
        tracker.waitForAll();
        m_fAllLoaded = !tracker.isErrorAny();
      }	catch (InterruptedException e) {
      }
			
      if (!m_fAllLoaded){
        stop();
        getGraphics().drawString("Error loading images!", 10, 40);
        return;
      }
			
      // Assuming all images are same width and height.
      m_imgWidth  = m_images[0].getWidth(this);
      m_imgHeight = m_images[0].getHeight(this);
      m_board.setImageSize(m_imgWidth,m_imgHeight);
    } else {
//      m_state = INTRO;
      repaint();
    }

    repaint();
    stop();
  }

}

